#define MAX(A,B) (((A) > (B)) ? (A) : (B))
#define ABS(A)   (((A) <  (0)) ? (-A): (A))

// solves 2-D Laplace equation using a relaxation scheme

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef _OPENACC
#include<openacc.h>
#endif

int main(int argc, char *argv[]) {
   double * T,  * Tnew; 
   double tol, var = DBL_MAX, top = 100.0;
   unsigned n, n2, maxIter, i, j, iter = 0;
   int itemsread;
   FILE *fout;

#include "../select_device.h"

#ifdef FIXED_INPUT
   n = 4096 ; maxIter = 100 ; tol = 0.0001;
#else
   printf("Enter mesh size, max iterations and tolerance: ");
   itemsread = scanf("%u ,%u ,%lf", &n, &maxIter, &tol);
   if (itemsread!=3) {
      fprintf(stderr, "Input error!\n");
      exit(-1);
   }
#endif


   n2 = n+2;
   T = (double*)calloc(n2*n2, sizeof(*T));
   Tnew = (double*)calloc(n2*n2, sizeof(*T));
   if (T == NULL || Tnew == NULL) {
      fprintf(stderr, "Not enough memory!\n");
      exit(EXIT_FAILURE);
   }

   // set boundary conditions

   for (i=1; i<=n; i++) {
      T[(n+1)*n2+i] =  i * top / (n+1);  
      T[i*n2+n+1] = i * top / (n+1);
   }
   for(i = 0; i<=n+1; i++) 
      for(j = 0; j<=n+1; j++) 
         Tnew[i*n2+j] = T[i*n2+j] ;

time_t startTime = clock();

#pragma acc data copy(T[0:n2*n2]), create(Tnew[0:n2*n2])
{
   while(var > tol && iter <= maxIter) {
      ++iter;
      var = 0.0;
#pragma acc parallel loop collapse(2) reduction(max:var)
      for (i=0; i<=n; ++i) {
         for (j=0; j<=n; ++j) {         
// work-around for PGI: start loops from 0 and use 'if' instead of starting from 1
            if(i>0 && j>0) {
                Tnew[i*n2+j] = 0.25*( T[(i-1)*n2+j] + T[(i+1)*n2+j] 
                                    + T[i*n2+(j-1)] + T[i*n2+(j+1)] );
        
#ifdef _OPENACC
                var = fmax(var, fabs(Tnew[i*n2+j] - T[i*n2+j]));
#else
                var = MAX(var, ABS(Tnew[i*n2+j] - T[i*n2+j]));
#endif
            }
         }
      }

#pragma acc parallel loop collapse(2)
      for (i=1; i<=n; ++i) {
         for (j=1; j<=n; ++j) {         
            T[i*n2+j] = Tnew[i*n2+j] ;
         }
      }
         if (iter%10 == 0)
            printf("iter: %8u, variation = %12.4lE\n", iter, var);
   }
}

   double endTime = (clock() - startTime) / (double) CLOCKS_PER_SEC;

   printf("Elapsed time (s) = %.2lf\n", endTime);
   printf("Mesh size: %u\n", n);
   printf("Stopped at iteration: %u\n", iter);
   printf("Maximum error: %lE\n", var);

#ifdef PRINT_OUTPUT
   // saving results to file
   fout = fopen("results", "w");
   if (fout == NULL) {
      perror("results");
      exit(-1);
   }
   for (i=1; i<=n; ++i)
      for (j=1; j<=n; ++j)
         fprintf(fout, "%8u %8u %18.9lE\n", i, j, T[i*n2+j]);
   fclose(fout);
#endif
   free(T);
   free(Tnew);
   return 0;
}
